/// gml_Alloc.h --
///
///   Memory allocation / deallocation for objects / structures.
//
//  Copyright (c) 1995-2004 CLIPS-IMAG
//
//  See the file "gml_LicenseTerms.txt" for information on usage and redistribution
//  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
//
//  Created on November, 1995 (FB).

#ifndef __GMLALLOC__
#define __GMLALLOC__


#include "gml_Types.h"
#include "gml_Errors.h"

#include <stdlib.h>
#include <string.h>


/// gml_Zero --
///
/// Set to zero all bytes in the data structure pointed to by <ptr>.

template<class T>
inline
void gml_Zero (T* ptr)
{
  memset ((void*)ptr, 0, sizeof (T));
}



/// gml_ZeroArray --
///
/// Set to zero all bytes of an array <ptr> that has <size> elements.

template<class T>
inline
void gml_ZeroArray (T* ptr, gml_TBlockSize size)
{
  memset ((void*)ptr, 0, sizeof (T) * size);
}



/// gml_Malloc --
///
/// Allocates an array of <size> structures.
/// On platforms other than windows, allocates page-aligned memory chunks if
/// the requested block is over half a page long.
///   <ptr>     the pointer in which the newly allocated address will be stored
///   <size>    the number of structures to allocate

template<class T>
inline
gml_TError gml_Malloc (T*& ptr, gml_TBlockSize size = 1)
{
  #ifdef WIN32
    ptr = (T*)malloc ((size_t)size * sizeof(T));

  #else
    if (size * sizeof(T) < 2048) {
      ptr = (T*)malloc ((size_t)size * sizeof(T));
    } else {
      ptr = (T*)valloc ((size_t)size * sizeof(T));
    }
  #endif

  if (ptr == (T*)NULL)
    return gml_gErrorMemFull;
  else
    return gml_cNoError;
}


/// gml_MallocAndZero --
///
/// Allocate an array of <size> structures, zero it, and stores its
/// pointer in <ptr>.
///
/// @see gml_Malloc() and gml_ZeroArray().

template<class T>
inline
gml_TError gml_MallocAndZero (T*& ptr, gml_TBlockSize size = 1)
{
  ptr = (T*)malloc ((size_t)size * sizeof(T));

  if (ptr == (T*)NULL)
    return gml_gErrorMemFull;
  else {
    gml_ZeroArray (ptr, size);
    return gml_cNoError;
  }
}




/// gml_Free --
///
/// Release resources claimed by gml_Malloc().
/// The pointer <ptr> will be set to NULL.

template<class T>
inline
void gml_Free (T*& ptr)
{
  free ((void*)ptr);
  ptr = (T*)NULL;
}




/// gml_TestAndFree --
///
/// Release resources claimed by gml_Malloc(), if <ptr> is not NULL.

template<class T>
inline
void gml_TestAndFree (T*& ptr)
{
  if (ptr != (T*)NULL)
    gml_Free (ptr);
}



/// gml_New --
///
/// Create a new instance of class <T> and stores a reference in <object>.

template<class T>
inline
gml_TError gml_New (T*& object)
{
  gml_TError err = gml_cNoError;

  object = new T;

  if (object == (T*)NULL)
    err = gml_gErrorMemFull;

  return err;
}



/// gml_NewArray --
///
/// Create a new array of <size> instances of class <T> and stores a reference in <object>.

template<class T>
inline
gml_TError gml_NewArray (T*& object, gml_TBlockSize size)
{
  gml_TError err = gml_cNoError;

  object = new T[size];

  if (object == (T*)NULL)
    err = gml_gErrorMemFull;

  return err;
}



/// gml_NewAndInit --
///
/// Create a new instance of class <T>, stores a reference in <object>
/// and call the object <Init> method.
/// The <Init> method is called without arguments.

template<class T>
inline
gml_TError gml_NewAndInit (T*& object)
{
  gml_TError err = gml_cNoError;

  err = gml_New (object);

  if (err == gml_cNoError) {
    err = object->Init ();
    if (err != gml_cNoError)
      gml_DisposeAndDelete (object);
  }

  return err;
}


/// gml_NewAndInitArray --
///
/// Create a new array of <size> instances of class <T>, stores a reference in
/// <object> and call the <Init> method on each instance.

template<class T>
inline
gml_TError gml_NewAndInitArray (T*& object, gml_TBlockSize size)
{
  gml_TError   err = gml_cNoError;
  int           i;

  if ((err = gml_NewArray (object, size)) != gml_cNoError)
    return err;

  for (i = 0; (err == gml_cNoError) && (i < size); i++)
    err = (object[i]).Init ();

  if (err != gml_cNoError) {
    for (; i != 0; i--)
      (object[i-1]).Dispose ();
    gml_DeleteArray (object);
  }

  return err;
}



/// gml_Delete --
///
/// Delete the instance <object> of class <T>. <object> is set to NULL.

template<class T>
inline
void gml_Delete (T*& object)
{
  delete object;
  object = (T*)NULL;
}



/// gml_DeleteArray --
///
/// Delete the array <array> of instances of class <T>. <array> is set to NULL.

template<class T>
inline
void gml_DeleteArray (T*& array)
{
  delete[] array;
  array = (T*)NULL;
}



/// gml_DisposeAndDelete --
///
/// Same as gml_Delete();
/// The object <Dispose> method will be called first.

template<class T>
inline
void gml_DisposeAndDelete (T*& object)
{
  object->Dispose ();

  gml_Delete (object);
}



/// gml_DisposeAndDeleteArray --
///
/// Same as gml_DeleteArray();
/// the objects <Dispose> method will be called first.

template<class T>
inline
void gml_DisposeAndDeleteArray (T*& array, gml_TBlockSize size)
{
  int i;
  
  for (i = 0; i < size; i++)
    (array[i]).Dispose ();

  gml_DeleteArray (array);
}



/// gml_TestDisposeAndDelete --
///
/// Same as gml_DisposeAndDelete();
/// nothing will be done if <object> is NULL.

template<class T>
inline
void gml_TestDisposeAndDelete (T*& object)
{
  if (object != (T*)NULL)
    gml_DisposeAndDelete (object);
}



/// gml_TestDisposeAndDeleteArray --
///
/// Same as gml_DisposeAndDeleteArray()
/// nothing will be done if <array> is NULL.

template<class T>
inline
void gml_TestDisposeAndDeleteArray (T*& array, gml_TBlockSize size)
{
  if (array != (T*)NULL)
    gml_DisposeAndDeleteArray (array, size);
}



/// gml_TestAndDelete --
///
/// Same as gml_Delete();
/// nothing will be done if <object> is NULL.

template<class T>
inline
void gml_TestAndDelete (T*& object)
{
  if (object != (T*)NULL)
    gml_Delete (object);
}



/// gml_TestAndDeleteArray --
///
/// Same as gml_DeleteArray();
/// nothing will be done if <array> is NULL.

template<class T>
inline
void gml_TestAndDeleteArray (T*& array)
{
  if (array != (T*)NULL)
    gml_DeleteArray (array);
}




/// gml_GetGlobalObject --
///
/// Put in <ref> a reference to the global object
/// of the class that <ref> is a pointer to. If no global object of this
/// class was allocated yet, do it.
///
/// Returns
///   gml_cNoError on success
///   gml_gErrorMemFull on failure, or an error code from the
///         <Init> initialisation method of the class.

template<class T>
inline
gml_TError gml_GetGlobalObject (T*& ref)
{
  gml_TError          err = gml_cNoError;

  // look if the global object have to be allocated
  if (T::fNumberOfRef == 0)
    // allocate a new object
    err = gml_NewAndInit (T::fGlobalRef);

  if (err == gml_cNoError) {
    ref = T::fGlobalRef;
    T::fNumberOfRef++;
  }

  return err;
}



/// gml_ReleaseGlobalObject --
///
/// If <ref> if not NULL, decrement the number
/// of references to the global object of the class that <ref> is a pointer to.
/// If this number becomes 0, dispose and delete the
/// global object. 
/// <ref> is set to NULL.

template<class T>
inline
void gml_ReleaseGlobalObject (T*& ref)
{
  if (ref != (T*)NULL) {

    if (T::fNumberOfRef == 1)
      gml_DisposeAndDelete (T::fGlobalRef);

    ref = (T*)NULL;

    T::fNumberOfRef--;
  }
}


#endif


