// gml_DynamicArray_T.h --
//
//    See "gmlDynamicArray.h".
//
//  Copyright (c) 1998-2004 CLIPS-IMAG
//
//  See the file "gml_LicenseTerms.txt" for information on usage and redistribution
//  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
//
//  Created on August 29, 1998 (FB).

#include "gml_DynamicArray.h"

#include "gml_Alloc.h"


static const int cWordBits     = 32;                // Number of bits per word in the bitmap
                                                    //  of used slots.
static const int cWordSize     = cWordBits / 8;     // Size (in bytes) of a word.
static const UInt32 cWordMask  = ~ 0;               // A word with all bits set.



// gml_TDynamicArray::Init --

template <class T>
gml_TError gml_TDynamicArray<T>::Init  (gml_TBlockSize    nbElem,
                                        gml_TBlockSize    nbIncrease)
{
  gml_TError   err         = gml_cNoError;

  // init data members
  fNbIncrease   = nbIncrease;

  fMem          = (T*)NULL;
  fNbElem       = 0;
  fInUse        = 0;

  fBitmap       = (UInt32*)NULL;
  fBMSize       = 0;

  err = this->Allocate (nbElem);

  if (err != gml_cNoError)
    this->Dispose ();

  return err;
}



// gml_TDynamicArray::Dispose --

template <class T>
void gml_TDynamicArray<T>::Dispose ()
{
  gml_TestAndFree (fMem);
  fNbElem = 0;
  fInUse  = 0;

  gml_TestAndFree (fBitmap);
  fBMSize = 0;
}



// gml_TDynamicArray::GetElem --

template <class T>
T* gml_TDynamicArray<T>::GetElem (gml_TBlockSize* index)
{
  gml_TError        err = gml_cNoError;

  int               i         = 0;
  gml_TBlockSize    j         = 0;
  gml_TBlockSize    theIndex;
  UInt32*           bmWord;
  UInt32            theWord;

  if (fInUse == fNbElem)
    err = this->Allocate (fNbElem + fNbIncrease);
  
  if (err != gml_cNoError)
    return (T*)NULL;

  bmWord = fBitmap;
  while ((j < fBMSize) && (*bmWord == cWordMask)) {
    bmWord++;
    j += 1;
  }

  theWord = *bmWord;
  while (theWord & ((UInt32)1 << i))
    i++;

  // Mark this slot as "used" in the bitmap

  *bmWord = theWord | ((UInt32)1 << i);
  fInUse ++;

  theIndex = (j * cWordBits + i);
  if (index != (gml_TBlockSize*)NULL)
    *index = theIndex;

  return fMem + theIndex;
}



// gml_TDynamicArray::ReleaseElem --

template <class T>
void gml_TDynamicArray<T>::ReleaseElem (T* elem)
{
  gml_TBlockSize  elemInd   = elem - fMem;
  UInt32          j         = elemInd / cWordBits;
  UInt32          i         = elemInd % cWordBits;

  fBitmap[j]  &= ~((UInt32)1 << i);
  fInUse --;
}




// gml_TDynamicArray::Allocate --

template <class T>
gml_TError gml_TDynamicArray<T>::Allocate (gml_TBlockSize nbElem)
{
  gml_TError        err;
  T*                mem         = (T*)NULL;
  UInt32*           bitmap      = (UInt32*)NULL;
  gml_TBlockSize    bmSize;
  int               div, mod;

  if (nbElem == 0)
    return gml_cNoError;

  // Allocate the new memory blocks for the array and the bitmap of used slots.
  
  if ((err = gml_Malloc (mem, nbElem)) != gml_cNoError)
    goto error;

  bmSize  = (nbElem - 1) / cWordBits + 1;
  if ((err = gml_Malloc (bitmap, bmSize)) != gml_cNoError)
    goto error;

  memset ((void*)bitmap, 0, bmSize * cWordSize);

  // If number of elements in the new array is not a multiple of
  //  the number of bits in a bitmap word, then set the last (unused)
  //  bits to "slot in use" (bit = 1).

  mod = nbElem % cWordBits;
  if (mod != 0)
    *(bitmap + (nbElem / cWordBits)) = (cWordMask << mod);

  // Copy the old values if necessary (i.e. if this is not the first allocation),
  //  and free the previous buffers.
  // We use memcpy because we know that the buffers don't overlap.

  if (fMem != (T*)NULL) {
    memcpy ((void*)mem, (void*)fMem, fNbElem * sizeof(T));

    div = fNbElem / cWordBits;
    mod = fNbElem % cWordBits;
    memcpy ((void*)bitmap, (void*)fBitmap, div * cWordSize);

    // If number of elements in the previous array is not a multiple of
    //  the number of bits in a bitmap word, then clear the unused
    //  bits while copying the last word.

    if (mod != 0)
      *(bitmap + div) |= ((*(fBitmap + div)) & (cWordMask >> (cWordBits - mod)));
    
    gml_Free (fMem);
    gml_Free (fBitmap);
  }

  fMem      = mem;
  fNbElem   = nbElem;

  fBitmap   = bitmap;
  fBMSize   = bmSize;

  return err;

error:
  // error handling
    gml_TestAndFree (mem);
    gml_TestAndFree (bitmap);

  return err;
}

