/// gml_List.h --
///
///    Definition of gml_TListElem and gml_TList.
///
//  Copyright (c) 2003-2004 CLIPS-IMAG
//
//  See the file "gml_LicenseTerms.txt" for information on usage and redistribution
//  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
//
//  Created on May 21, 2003 (FB).

#ifndef __GML_LIST__
#define __GML_LIST__

#include "gml_Types.h"
#include "gml_Errors.h"



/// gml_TListSortFunc --
///   The prototype of functions that compare two elements <left> and <right>,
///   and return an integer that is greater than, equal to, or less than 0, according as
///   <left> is greater than, equal to, or less than <right>.

typedef int (gml_TListSortFunc) (void* left, void* right);


/// gml_ListSortStringFunc --
///
///  Element-sorting routine that compares C (0-terminated) strings using strcmp.

extern int gml_ListSortStringFunc (void* left, void* right);

/// gml_ListSortIntFunc --
///
///  Element-sorting routine that compares signed integers.

extern int gml_ListSortIntFunc (void* left, void* right);


class gml_TList;

/// gml_TListElem --
///
///  An element of a gml_TList.

class gml_TListElem
{
  public:

    /// Zero all fields.
    virtual gml_TError  Init ();
    
    /// Call RemoveFromList()
    virtual void        Dispose ();

    /// If this element is inside a list, remove it from the list
    inline void         RemoveFromList ();
    inline void         Reset ();

    /// Return the next element in the list (possibly NULL).
    gml_TListElem*      Next ()
                          { return fNext; }

    /// Return the previous element in the list (possibly NULL).
    gml_TListElem*      Previous ()
                          { return fPrevious; }

    /// Return the element's associated value.
    void*               Value ()
                          { return fValue; }

    /// The element's associated value.
    void*               fValue;

    virtual             ~gml_TListElem () {};
    // please the compiler

  protected:

    friend class gml_TList;
  
    gml_TList*          fList;      ///< pointer to the containing list
    gml_TListElem*      fNext;      ///< pointer to the next list element
    gml_TListElem*      fPrevious;  ///< pointer to the previous list element
};



/// gml_TList --
///   A list of gml_TListElem elements.

class gml_TList
{
  public:

    /// Init --
    ///
    ///   Initializes the list to an empty state.

    inline gml_TError         Init ();

    /// Dispose --
    ///
    ///   Disposes the list resources, which includes disposing and deleting all
    ///   the list elements.

    void                      Dispose ();

    /// Reset --
    ///
    ///   Resets the list to an empty state. Does not affect the list elements that
    ///   where in the list before the call.

    inline void               Reset ();

    /// Head --
    ///
    ///   Returns the first element in the list.

    gml_TListElem*            Head ()
                                { return fHead; }

    /// Tail --
    ///
    ///   Returns the last element in the list.

    gml_TListElem*            Tail ()
                                { return fTail; }
    
    /// NbElem --
    ///
    ///   Returns the number of elements in the list.

    inline int                NbElem ();
    
    /// AppendElem --
    ///
    ///   Inserts <elem> at the end of the list. <elem> should not be NULL.

    inline void               AppendElem    (gml_TListElem* elem);

    /// InsertElem --
    ///
    ///   Inserts <elem> at the beginning of the list. <elem> should not be NULL.

    inline void               InsertElem    (gml_TListElem* elem);

    /// RemoveElem --
    ///
    ///   Removes <elem> from the list. <elem> should not be NULL.

    inline void               RemoveElem    (gml_TListElem* elem);

    /// RemoveElem --
    ///
    ///   Checks if the element <elem> is a member of the list.
    ///   Returns <gml_cTrue> if <elem> is in the list, <gml_cFalse> otherwise.
    ///   <elem> should not be NULL.

    inline gml_TBoolean       IsElemMember  (gml_TListElem* elem);

    /// InsertListAfterElem --
    ///
    ///   Inserts the elements of <list> after <elem>.
    ///   If <elem> is NULL, insert <list> elements at the beginning of this list.
    ///   Does nothing if <elem> is not an element of the list.
    ///   Doesn't allocate any element, just change <list> elements to belong to this list.
    ///   Finally, <list> is reset (to an empty list).

    void                      InsertListAfterElem (gml_TList* list, gml_TListElem* elem);



    /// AppendValue --
    ///
    ///   Allocates a new element and append it to the list.
    ///   Set the value of the element to <value>.
    ///   If <newElem> is not NULL, it will receive the address of the newly allocated element.

    gml_TError                AppendValue   (void* value, gml_TListElem** newElem = (gml_TListElem**)NULL);

    /// InsertValue --
    ///
    ///   Allocates a new element and insert it at the beginning of the list. 
    ///   Set the value of the element to <value>.
    ///   If <newElem> is not NULL, it will receive the address of the newly allocated element.

    gml_TError                InsertValue   (void* value, gml_TListElem** newElem = (gml_TListElem**)NULL);

    /// ValueElem --
    ///
    ///   Returns a pointer to the first element in the list which value is <value>, or NULL if no
    ///   element have such a value.

    gml_TListElem*            ValueElem     (void* value);

    /// AppendValuetAfterElem --
    ///
    ///   Allocates a new element and insert it in the list just after <elem>.
    ///   If <elem> is NULL, insert the new element at the head of the list.
    ///   Set the value of the element to <value>.
    ///   If <newElem> is not NULL, it will receive the address of the newly allocated element.
    
    gml_TError                AppendValuetAfterElem (void* value, gml_TListElem* elem, gml_TListElem** newElem = (gml_TListElem**)NULL);
    
    /// RemoveValue --
    ///
    ///   Finds the first element in the list that has the value <value>, removes it from the list
    ///   and free it. Do nothing if no element of the list have value <value>.

    void                      RemoveValue   (void* value);

    /// IsValueMember --
    ///
    ///   Returns "gml_cTrue" if at least one element in list has the value <value>.

    inline gml_TBoolean       IsValueMember (void* value);

    /// IsEquivalentValueMember --
    ///
    ///   Returns "gml_cTrue" if at least one element in list has a value that is equivalent to
    ///   <value>. The equivalence function must be provided in <func>. By default, <func> is
    ///   the C-string equivalent functions (2 values are equivalent if they point to C-strings
    ///   that have the same characters).

    gml_TBoolean              IsEquivalentValueMember (void* value,
                                                        gml_TListSortFunc* func = &gml_ListSortStringFunc);

    /// Sort --
    ///
    ///   Sorts the list according to the <func> comparison function.

    gml_TError                Sort (gml_TListSortFunc* func = &gml_ListSortStringFunc);

  protected:

    gml_TListElem*      fHead;  ///< the first element in this list
    gml_TListElem*      fTail;  ///< the last element in this list
};



// gml_TListElem Inline function definitions.

// gml_TListElem::RemoveFromList --

void gml_TListElem::RemoveFromList ()
{
  if (fList != (gml_TList*)NULL)
    fList->RemoveElem (this);
}



// gml_TListElem::Reset --

void gml_TListElem::Reset ()
{
  fNext         = (gml_TListElem*)NULL;
  fPrevious     = (gml_TListElem*)NULL;
  fList         = (gml_TList*)NULL;
}



// gml_TList Inline function definitions.

// gml_TList::Init --

gml_TError gml_TList::Init ()
{
  this->Reset ();

  return gml_cNoError;
}

// gml_TList::Reset --

void gml_TList::Reset ()
{
  fHead = (gml_TListElem*)NULL;
  fTail = (gml_TListElem*)NULL;
}



// gml_TList::NbElem --

int gml_TList::NbElem ()
{
  int               nb      = 0;
  gml_TListElem*    elem    = fHead;

  while (elem != (gml_TListElem*)NULL) {
    nb++;
    elem            = elem->fNext;
  }

  return nb;
}



// gml_TList::AppendElem --

void gml_TList::AppendElem (gml_TListElem* elem)
{
  elem->fList       = this;
  elem->fPrevious   = fTail;
  elem->fNext       = (gml_TListElem*)NULL;
  
  if (fTail != (gml_TListElem*)NULL)
    fTail->fNext    = elem;
  else
    fHead           = elem;
  fTail             = elem;
}



// gml_TList::InsertElem --

void gml_TList::InsertElem (gml_TListElem* elem)
{
  elem->fList         = this;
  elem->fNext         = fHead;
  elem->fPrevious     = (gml_TListElem*)NULL;
  
  if (fHead != (gml_TListElem*)NULL)
    fHead->fPrevious  = elem;
  else
    fTail             = elem;
  fHead               = elem;
}

// gml_TList::RemoveElem --

void gml_TList::RemoveElem (gml_TListElem* elem)
{
  if (elem->fList != this)
    return;

  if (elem != fHead)
    elem->fPrevious->fNext = elem->fNext;
  else
    fHead = elem->fNext;

  if (elem != fTail)
    elem->fNext->fPrevious = elem->fPrevious;
  else
    fTail         = elem->fPrevious;

  elem->Reset ();
}



// gml_TList::IsElemMember --

gml_TBoolean gml_TList::IsElemMember (gml_TListElem* elem)
{
  return (gml_TBoolean)(elem->fList == this);
}



// gml_TList::IsValueMember --

gml_TBoolean gml_TList::IsValueMember (void* value)
{
  return (gml_TBoolean)((this->ValueElem (value) != (gml_TListElem*)NULL));
}




#endif
