// gmlTcl_Class.cpp --
//
//    Defines the gmlTcl_TClass, see "gmlTcl_Class.h".
//
//  Copyright (c) 2007 LIG/IIHM
//
//  See the file "gml_LicenseTerms.txt" for information on usage and redistribution
//  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
//
//  Created on June 21, 2007 (FB).

#include "gmlTcl_Class.h"

extern "C" {
  #include "tcl.h"
}

#include "gmlTcl_Obj_Internal.h"

#include "gmlTcl_Entity.h"
#include "gmlTcl_Object.h"
#include "gmlTcl_Method.h"

#include "gml_Alloc.h"






// gmlTcl_TClass::Init --

int gmlTcl_TClass::Init (Tcl_Interp* interp, Tcl_Obj* name)
{
  int                 res           = TCL_OK;
  gml_TError          err           = gml_cNoError;

  fObjectList         = (gml_TList*)NULL;
  fMethodTableInit    = 0;
  fConstructor        = (gmlTcl_TMethod*)NULL;
  fDestructor         = (gmlTcl_TMethod*)NULL;
  fAncestors          = (gml_TList*)NULL;
  fSpecializations    = (gml_TList*)NULL;

  // Call superclass initialization
  
  res = this->gmlTcl_TEntity::Init (gmlTcl_kClassEntityType, interp, name);
  if (res != TCL_OK)
    return res;

  // Initialize the method hash table and the objects, ancestors and specializations lists.
  
  Tcl_InitHashTable (&fMethodTable, TCL_STRING_KEYS);
  fMethodTableInit = 1;

  err = gml_NewAndInit (fObjectList);
  if (err == gml_cNoError)
    err = gml_NewAndInit (fAncestors);
  if (err == gml_cNoError)
    err = gml_NewAndInit (fSpecializations);

  if (err != gml_cNoError) {
    Tcl_AppendResult (interp, "not enough memory", (char*)NULL);
    res = TCL_ERROR;
    goto error;
  }

  // The first ever class created is the root class.
  // All other classes inherit from the root class.

  if (gmlTcl_gObj_RootClass == (gmlTcl_TClass*)NULL)
    gmlTcl_gObj_RootClass = this;
  else
    res = this->Inherit (interp, gmlTcl_gObj_RootClass);

  fState = gmlTcl_kEntityState_normal;
  return res;

error:
  this->Dispose ();
  return TCL_ERROR;
}



// gmlTcl_TClass::Dispose --

void gmlTcl_TClass::Dispose ()
{
  Tcl_HashEntry*      entry;
  Tcl_HashSearch      search;
  gmlTcl_TMethod*     theMethod;
  gml_TListElem*      elem;

  // Remove the global variable of this class. Only do it if the class was
  //  initialized ok, otherwise we may be deleting a global variable that existed
  //  before this entity creation attempt and that caused the dispose.

  if (fAncestors != (gml_TList*)NULL)
    Tcl_UnsetVar (gmlTcl_Obj_Interp (), fStringName, TCL_GLOBAL_ONLY);

  // De-inherit this class from its ancestors.
  // De-inherit this class specializations from it.
  // Free the ancestors and specializations.
  
  if (fAncestors != (gml_TList*)NULL) {
    while ((elem = fAncestors->Head ()) != (gml_TListElem*)NULL)
      this->DeInherit ((gmlTcl_TClass*)(elem->Value ()));
    gml_DisposeAndDelete (fAncestors);
  }
   
  if (fSpecializations != (gml_TList*)NULL) {
    while ((elem = fSpecializations->Head ()) != (gml_TListElem*)NULL)
      ((gmlTcl_TClass*)(elem->Value ()))->DeInherit (this);
    gml_DisposeAndDelete (fSpecializations);
  }

  // Free the object list, freeing all objects of this class.

  if (fObjectList != (gml_TList*)NULL) {
    while ((elem = fObjectList->Head ()) != (gml_TListElem*)NULL)
      ((gmlTcl_TObject*)(elem->Value ()))->Dispose ();

    gml_DisposeAndDelete (fObjectList);
  }

  // Free the method hash table

  if (fMethodTableInit) {

    // Free all methods of this class

    while ((entry = Tcl_FirstHashEntry (&fMethodTable, &search)) != (Tcl_HashEntry*)NULL) {
      theMethod = (gmlTcl_TMethod*)Tcl_GetHashValue (entry);
      theMethod->Dispose ();
      delete theMethod;
    }
    Tcl_DeleteHashTable (&fMethodTable);
    fMethodTableInit = 0;
  }

  // Call the superclass disposal.

  this->gmlTcl_TEntity::Dispose ();
}



// CreateRecursiveAncestorList --

gml_TError gmlTcl_TClass::CreateRecursiveAncestorList (gml_TList** list)
{
  gml_TError          err         = gml_cNoError;
  gml_TList*          theList     = (gml_TList*)NULL;
  gml_TListElem*      elem        = fAncestors->Head ();
  gmlTcl_TClass*      ancestor    = (gmlTcl_TClass*)NULL;

  // If *list is NULL, allocate a new list, otherwise use the provided list.

  if (*list == (gml_TList*)NULL)
    err = gml_NewAndInit (theList);
  else
    theList = *list;

  while ((err == gml_cNoError) && (elem != (gml_TListElem*)NULL)) {
    ancestor = (gmlTcl_TClass*)(elem->Value ());
    if (!(theList->IsValueMember ((void*)ancestor))) {
      err = theList->AppendValue ((void*)ancestor);
      if (err == gml_cNoError)
        err = ancestor->CreateRecursiveAncestorList (&theList);
    }
    if (err == gml_cNoError)
      elem = elem->Next ();
  }

  if (*list == (gml_TList*)NULL)
    if (err != gml_cNoError)
      gml_TestDisposeAndDelete (theList);
    else
      *list = theList;

  return err;
}



// IsAncestor --

int gmlTcl_TClass::IsAncestor (gmlTcl_TClass* ancestor)
{
  gml_TListElem*      elem        = fAncestors->Head ();
  gmlTcl_TClass*      anAncestor;
  
  while (elem != (gml_TListElem*)NULL) {
    anAncestor = (gmlTcl_TClass*)(elem->Value ());
    if (anAncestor == ancestor)
      return 1;
    if (anAncestor->IsAncestor (ancestor))
      return 1;

    elem = elem->Next ();
  }
  return 0;
}



// Inherit --

int gmlTcl_TClass::Inherit (Tcl_Interp* interp, gmlTcl_TClass* superClass)
{
  gml_TError      err;
  gml_TListElem*  ancestorElem          = (gml_TListElem*)NULL;
  gml_TListElem*  specializationElem    = (gml_TListElem*)NULL;

  err = gml_NewAndInit (ancestorElem);
  if (err == gml_cNoError)
    err = gml_NewAndInit (specializationElem);

  if (err != gml_cNoError) {
    Tcl_AppendResult (interp, "not enough memory", (char*)NULL);
    gml_TestDisposeAndDelete (ancestorElem);
    return TCL_ERROR;
  }

  ancestorElem->fValue = (void*)superClass;
  fAncestors->InsertElem (ancestorElem);

  specializationElem->fValue = (void*)this;
  superClass->fSpecializations->AppendElem (specializationElem);

  return TCL_OK;
}



// DeInherit --

void gmlTcl_TClass::DeInherit (gmlTcl_TClass* superClass)
{
  gml_TListElem*      elem;

  elem = (gml_TListElem*)(fAncestors->ValueElem ((void*)superClass));
  gml_TestDisposeAndDelete (elem);

  elem = (gml_TListElem*)(superClass->fSpecializations->ValueElem ((void*)this));
  gml_TestDisposeAndDelete (elem);
}



// CommandHandler --
//
//  Creates new objects of this class.

int gmlTcl_TClass::CommandHandler  (Tcl_Interp*   interp,
                                    int           objc,
                                    Tcl_Obj*      objv[])
{
  int               res         = TCL_OK;
  gmlTcl_TObject*   theObject   = (gmlTcl_TObject*)NULL;

  // Check arguments
  
  if (objc < 2) {
    Tcl_AppendResult (interp, "wrong # args: should be \"",
                              fProtectedString, " <objName> ...\"", (char*)NULL);
    return TCL_ERROR;
  }

  // Create the object

  theObject = new gmlTcl_TObject;
  if (theObject == (gmlTcl_TObject*)NULL) {
    Tcl_AppendResult (interp, "could not create object \"",
                              Tcl_GetString (objv[1]), "\": not enough memory",
                              (char*)NULL);
    return TCL_ERROR;
  }
  res = theObject->Init (interp, this, objc, objv);
  if (res != TCL_OK) {
    theObject->Dispose ();
    delete theObject;
  }

  if (res == TCL_OK)
    Tcl_SetObjResult (interp, objv[1]);

  return res;
}





// RegisterObject --

void gmlTcl_TClass::RegisterObject (gmlTcl_TObject* theObject)
{
  gml_TError        err;

  err = fObjectList->AppendValue ((void*)theObject);
  if (err != gml_cNoError)
    fprintf (stderr, "PANIC: can't register an instance to its class: no more memory\n");
}



// DeregisterObject --

void gmlTcl_TClass::DeregisterObject (gmlTcl_TObject* theObject)
{
  fObjectList->RemoveValue   ((void*)theObject);
}



// RegisterMethod --

void gmlTcl_TClass::RegisterMethod (gmlTcl_TMethod* theMethod, int flags)
{
  Tcl_HashEntry*    entry;
  gmlTcl_TMethod*   aMethod;
  int               newEntry;

  entry = Tcl_CreateHashEntry (&fMethodTable, theMethod->StringName (), &newEntry);

  if (!newEntry) {
    // A method with this name already existed, destroy it.
    //  Dispose will destroy the entry in the hash table, so re-create it.

    aMethod = (gmlTcl_TMethod*)Tcl_GetHashValue (entry);
    aMethod->Dispose ();
    delete aMethod;

    entry = Tcl_CreateHashEntry (&fMethodTable, theMethod->StringName (), &newEntry);
  }

  Tcl_SetHashValue (entry, (ClientData)theMethod);

  if (flags & 1)
    fConstructor = theMethod;
  else if (flags & 2)
    fDestructor = theMethod;
}



// DeregisterMethod --

void gmlTcl_TClass::DeregisterMethod (gmlTcl_TMethod* theMethod, int flags)
{
  Tcl_HashEntry*    entry;

  entry = Tcl_FindHashEntry (&fMethodTable, theMethod->StringName ());
  if (entry == (Tcl_HashEntry*)NULL) {
    fprintf (stderr, "PANIC: can't deregister a method from its class when its not in the class's method table\n");
    return;
  }

  Tcl_DeleteHashEntry (entry);

  if (flags & 1)
    fConstructor = (gmlTcl_TMethod*)NULL;
  else if (flags & 2)
    fDestructor = (gmlTcl_TMethod*)NULL;
}



// FindMethod --

gmlTcl_TMethod* gmlTcl_TClass::FindMethod  (Tcl_Obj*          methodNameObj,
                                            Tcl_Interp*       interp,
                                            int               flags)
{
  gml_TError        err           = gml_cNoError;
  Tcl_HashEntry*    entry;
  gmlTcl_TMethod*   method        = (gmlTcl_TMethod*)NULL;
  char*             methodStr     = Tcl_GetString (methodNameObj);
  gml_TListElem*    elem;
  gmlTcl_TClass*    ancestor;
  gml_TList*        methodList;

  if (flags & 1) {
    // Lookup in this class method table.
  
    entry = Tcl_FindHashEntry (&fMethodTable, methodStr);
    if (entry != (Tcl_HashEntry*)NULL)
      return (gmlTcl_TMethod*)Tcl_GetHashValue (entry);
  }

  // Lookup in this class ancestors.

  if (flags & 2) {
    elem = fAncestors->Head ();
    while ((method == (gmlTcl_TMethod*)NULL) && (elem != (gml_TListElem*)NULL)) {
      ancestor = (gmlTcl_TClass*)(elem->Value ());
      method = ancestor->FindMethod (methodNameObj);
      elem = elem->Next ();
    }
  }

  // Create an error message if requested (i.e. interp != NULL)

  if ((interp != (Tcl_Interp*)NULL) && (method == (gmlTcl_TMethod*)NULL))
    if (flags & 1) {

      // The search started in this class methods, this was a regular search, build the list
      //  of callable methods on objects of this class.

      err = this->FullMethodList (&methodList, 0);

      if (err != gml_cNoError)
        Tcl_AppendResult (interp, "wrong method \"", methodStr, "\"", (char*)NULL);
      else {
        if (methodList->Head () == (gml_TListElem*)NULL)
          Tcl_AppendResult (interp, "wrong method \"", methodStr, "\": class \"", 
                                    fStringName, "\" has no methods\"", (char*)NULL);
        else {
          Tcl_AppendResult (interp, "wrong method \"", methodStr, "\": should be ", (char*)NULL);
          gmlTcl_Obj_AppendOptionList (interp, methodList);
        }

        while ((elem = methodList->Head ()) != (gml_TListElem*)NULL) {
          free ((void*)(elem->Value ()));
          gml_DisposeAndDelete (elem);
        }
        gml_DisposeAndDelete (methodList);
      }

    } else
      // The was restricted to the class superclasses, this was an inherited::* call.
      Tcl_AppendResult (interp, "\"", fStringName, "\" does not specialize the method \"",
                                methodStr, "\"", (char*)NULL);
  return method;
}



// FullMethodList --

gml_TError gmlTcl_TClass::FullMethodList (gml_TList** list, int superPrefix)
{
  gml_TError          err           = gml_cNoError;
  gml_TList*          result        = (gml_TList*)NULL;
  gml_TList*          alreadyIn     = (gml_TList*)NULL;
  gml_TListElem*      elem          = (gml_TListElem*)NULL;

  err = gml_NewAndInit (result);
  if (err == gml_cNoError)
    err = gml_NewAndInit (alreadyIn);

  if (err == gml_cNoError)
    err = this->AddMethodsToList (alreadyIn, result, 0, 0);

  if (err == gml_cNoError) {
    elem = fAncestors->Head ();
    while ((err == gml_cNoError) && (elem != (gml_TListElem*)NULL)) {
      err = ((gmlTcl_TClass*)(elem->Value ()))->AddMethodsToList (alreadyIn, result, superPrefix, 1);
      if (err == gml_cNoError)
        elem = elem->Next ();
    }
  }

  // Free the alreadyIn list.

  if (alreadyIn != (gml_TList*)NULL) {
    while ((elem = alreadyIn->Head ()) != (gml_TListElem*)NULL) {
      free ((void*)(elem->Value ()));
      gml_DisposeAndDelete (elem);
    }
    gml_DisposeAndDelete (alreadyIn);
  }

  if (!superPrefix)
    err = result->Sort ();  

  if (err != gml_cNoError) {

    // Free the result list and report error.

    if (result != (gml_TList*)NULL) {
      while ((elem = result->Head ()) != (gml_TListElem*)NULL) {
        free ((void*)(elem->Value ()));
        gml_DisposeAndDelete (elem);
      }
      gml_DisposeAndDelete (result);
    }
  }

  *list = result;

  return err;
}



// AddMethodsToList --

gml_TError gmlTcl_TClass::AddMethodsToList (gml_TList*          alreadyIn,
                                            gml_TList*          result,
                                            int                 prefix,
                                            int                 recursive)
{
  gml_TError          err           = gml_cNoError;
  Tcl_HashEntry*      entry         = (Tcl_HashEntry*)NULL;
  Tcl_HashSearch      search;
  gml_TListElem*      elem          = (gml_TListElem*)NULL;
  char*               methodStr;
  char*               methodBuffer;
  int                 unprefixedLen;
  int                 resultLen;

  // Add all methods defined by this class in the list when they are not already
  //  in the list.

  entry = Tcl_FirstHashEntry (&fMethodTable, &search);
  while ((err == gml_cNoError) && (entry != (Tcl_HashEntry*)NULL)) {
    methodStr           = ((gmlTcl_TMethod*)Tcl_GetHashValue (entry))->StringName ();

    if (!(alreadyIn->IsEquivalentValueMember ((char*)methodStr))) {

      unprefixedLen     = strlen (methodStr) + 1;
      resultLen         = prefix ? (strlen (fStringName) + 2 +strlen (methodStr) + 1)
                                   : unprefixedLen;

      methodBuffer = (char*)malloc (unprefixedLen);
      if (methodBuffer == (char*)NULL)
        err = gml_gErrorMemFull;
      else {
        strcpy (methodBuffer, methodStr);
        err = alreadyIn->AppendValue ((void*)methodBuffer);
      }

      // Add the methodName prefixed with the className in the prefixed list if requestsed,
      //  otherwise add the unprefixed methodName.

      if (err == gml_cNoError) {
        methodBuffer = (char*)malloc (resultLen);
        if (methodBuffer == (char*)NULL)
          err = gml_gErrorMemFull;
        else
          if (prefix)
            sprintf (methodBuffer, "%s::%s", fStringName, methodStr);
          else
            strcpy (methodBuffer, methodStr);

        err = result->AppendValue ((void*)methodBuffer);
      }
    }

    if (err == gml_cNoError)
      entry = Tcl_NextHashEntry (&search);
  }
  if (err != gml_cNoError)
    return err;

  // Add methods from superclasses if requested.
  
  if (recursive) {
    elem = fAncestors->Head ();
    while ((err == gml_cNoError) && (elem != (gml_TListElem*)NULL)) {
      err = ((gmlTcl_TClass*)(elem->Value ()))->AddMethodsToList (alreadyIn, result, prefix, recursive);
      if (err == gml_cNoError)
        elem = elem->Next ();
    }
    if (err != gml_cNoError)
      return err;
  }

  return err;
}
