// gmlTcl_Entity.cpp --
//
//    Defines the gmlTcl_TEntity, see "gmlTcl_Entity.h".
//
//  Copyright (c) 2007 LIG/IIHM
//
//  See the file "gml_LicenseTerms.txt" for information on usage and redistribution
//  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
//
//  Created on July 20, 2007 (FB).

#include "gmlTcl_Entity.h"

extern "C" {
  #include <stdlib.h>
  #include <string.h>

  #include "tclInt.h"
  #include "tclCompile.h"
}

#include "gml_Alloc.h"

#include "gmlTcl_Class.h"
#include "gmlTcl_Object.h"


extern "C"
int gmlTcl_EntityCmdHandler (ClientData    clientData,
                             Tcl_Interp*   interp,
                             int           objc,
                             Tcl_Obj*      objv[]);

extern "C"
void gmlTcl_EntityDeleteCmdHandler (ClientData clientData);

extern "C"
void gmlTcl_EntityRenameCallback (ClientData  clientData,
                                  Tcl_Interp* interp,
                                  char*       oldName,
                                  char*       newName,
                                  int         flags);





// gmlTcl_TEntity::Init --

int gmlTcl_TEntity::Init (gmlTcl_TEntityType type, Tcl_Interp* interp, Tcl_Obj* name)
{
  int                 res               = TCL_OK;
  int                 varExists;
  Tcl_HashEntry*      entry;
  Namespace*          globalNsPtr       = (Namespace*)Tcl_GetGlobalNamespace (interp);

  fState              = gmlTcl_kEntityState_inInit;
  fType               = type;
  fTypeString         = gmlTcl_kEntityTypeStrings[fType];
  fTable              = gmlTcl_Obj_EntityHashTable (this);
  fInTable            = 0;
  fStringName         = (char*)NULL;
  fProtectedString    = (char*)NULL;
  fCmd                = (Tcl_Command)NULL;

  // Create the string names.

  res = gmlTcl_Obj_UpdateStringNames (Tcl_GetString (name), &fStringName, &fProtectedString, interp);
  if (res != TCL_OK)
    goto error;
  
  // Only create a new entity if there is no entity, no global variable and no procedure with
  //    the entity name.

  if (gmlTcl_Obj_EntityFromTclObj (fType, name) != (gmlTcl_TEntity*)NULL) {
    Tcl_AppendResult (interp, "could not create ", fTypeString, " \"", fStringName,
                              "\": a ", fTypeString, " with that name already exists",
                              (char*)NULL);
    goto error;
  }

  res = gmlTcl_Obj_GlobalExists (interp, fStringName, &varExists);
  if (res != TCL_OK)
    goto error;

  if (varExists) {
    Tcl_AppendResult (interp, "could not create ", fTypeString, " \"", fStringName,
                              "\": a global variable with that name already exists",
                              (char*)NULL);
    goto error;
  }

  entry = Tcl_FindHashEntry (&globalNsPtr->cmdTable, fStringName);
  if (entry != (Tcl_HashEntry*)NULL) {
    Tcl_AppendResult (interp, "could not create ", fTypeString, " \"", fStringName,
                              "\": a command with that name already exists",
                              (char*)NULL);
    goto error;
  }

  // Create the command for the entity. This command has the same name as the entity.

  fCmd = Tcl_CreateObjCommand (interp, fStringName, gmlTcl_EntityCmdHandler, (ClientData)this,
                                gmlTcl_EntityDeleteCmdHandler);
  if (fCmd == (Tcl_Command)NULL) {
    Tcl_AppendResult (interp, "could not create ", fTypeString, " \"", fStringName,
                              "\": failed to create the ", fTypeString, " command",
                              (char*)NULL);
    goto error;
  }

  // Register this entity its global hash table (string -> pointer)

  gmlTcl_Obj_RegisterEntity (this);
  fInTable = 1;

  // Hook the callback on object renaming

  res = Tcl_TraceCommand (interp, fStringName, TCL_TRACE_RENAME,
                          gmlTcl_EntityRenameCallback, (ClientData)this);

  return res;

error:
  this->Dispose ();
  return TCL_ERROR;
}



// gmlTcl_TClass::Dispose --

void gmlTcl_TEntity::Dispose ()
{
  fState = gmlTcl_kEntityState_inDispose;

  // Remove the trace on renaming the command.

  if (fStringName != (char*)NULL) {
    Tcl_UntraceCommand (gmlTcl_Obj_Interp (), fStringName, TCL_TRACE_RENAME,
                        gmlTcl_EntityRenameCallback, (ClientData)this);
  }

  // Remove this entity command from the interpreter

  if (fCmd != (Tcl_Command)NULL) {
    Tcl_DeleteCommandFromToken (gmlTcl_Obj_Interp (), fCmd);
    fCmd = (Tcl_Command)NULL;
  }

  // De-register this entity from the global hash table (string -> pointer)

  if (fInTable) {
    gmlTcl_Obj_DeregisterEntity (this);
    fInTable = 0;
  }

  // Free the string representations of this class name

  gml_TestAndFree (fProtectedString);
  gml_TestAndFree (fStringName);
}






// CommandDeleted --

void gmlTcl_TEntity::CommandDeleted ()
{

  switch (fState) {
    case gmlTcl_kEntityState_normal:
      // Command is deleted by an event external to this entity, Dispose and delete the entity.

      this->Dispose ();
      delete this;
      break;

    case gmlTcl_kEntityState_inInit:
      // Command is deleted while the object is being initialized, i.e. during constructor evaluation.
      //  Flag this so it is handled in the initialization.

      fState = gmlTcl_kEntityState_cmdDeleted;
      break;

    case gmlTcl_kEntityState_inDispose:
      // Command is deleted as part of the normal entity disposal, do nothing.
      break;
  }
}



// Rename --

void gmlTcl_TEntity::Rename  (Tcl_Interp* interp,
                              char*       oldName,
                              char*       newName,
                              int         /* flags */)
{
  int                 res           = TCL_OK;
  int                 varExists;
  int                 oldNameLen    = strlen (oldName);
  int                 newNameLen    = strlen (newName);
  Tcl_HashEntry*      entry;
  int                 newEntry;

  // Cleanup any "::" header in front of the names.
  
  if ((oldNameLen > 2) && (oldName[0] = ':') && (oldName[1] = ':')) {
    oldName      += 2;
    oldNameLen   -= 2;
  }
  if ((newNameLen > 2) && (newName[0] = ':') && (newName[1] = ':')) {
    newName      += 2;
    newNameLen   -= 2;
  }

  // Remove the trace on the old name.

  Tcl_UntraceCommand (interp, fStringName, TCL_TRACE_RENAME,
                      gmlTcl_EntityRenameCallback, (ClientData)this);

  // If a global variable with the new name already exists, unset it

  res = gmlTcl_Obj_GlobalExists (interp, newName, &varExists);
  if (res != TCL_OK)
    return;

  if (varExists)
    Tcl_UnsetVar (interp, newName, TCL_GLOBAL_ONLY);

  // If a global variable with the old name already exists, copy its content to a
  //  global variable with the new name and unset the old global.

  res = gmlTcl_Obj_GlobalExists (interp, oldName, &varExists);
  if (res != TCL_OK)
    return;

  if (varExists) {

    res = gmlTcl_Obj_CloneGlobalArray (interp, oldName, newName);
    if (res != TCL_OK)
      return;

    Tcl_UnsetVar (interp, oldName, TCL_GLOBAL_ONLY);
  }

  // Update the entry key in the global string to pointer hash table.

  entry = Tcl_FindHashEntry (fTable, fStringName);
  if (entry == (Tcl_HashEntry*)NULL) {
    fprintf (stderr, "PANIC: entry not found in the the string to pointer hash entry while renaming %s \"%s\" to \"%s\"\n", fTypeString, fStringName, newName);
    return;
  }
  Tcl_DeleteHashEntry (entry);
  entry = Tcl_CreateHashEntry (fTable, newName, &newEntry);
  if (!newEntry) {
    fprintf (stderr, "PANIC: new name already exist in the the string to pointer hash entry while renaming %s \"%s\" to \"%s\"\n", fTypeString, fStringName, newName);
    return;
  }
  Tcl_SetHashValue (entry, (ClientData)this);

  // Update the string name

  res = gmlTcl_Obj_UpdateStringNames (newName, &fStringName, &fProtectedString);
  if (res != TCL_OK) {
    fprintf (stderr, "PANIC: no more memory while renaming %s \"%s\" to \"%s\"\n", fTypeString, fStringName, newName);
    return;
  }

  // Create trace on the new name
  
  Tcl_TraceCommand (interp, fStringName, TCL_TRACE_RENAME,
                    gmlTcl_EntityRenameCallback, (ClientData)this);
}






////////// Global (C) functions


// gmlTcl_EntityCmdHandler --

extern "C"
int gmlTcl_EntityCmdHandler (ClientData    clientData,
                             Tcl_Interp*   interp,
                             int           objc,
                             Tcl_Obj*      objv[])
{
  return ((gmlTcl_TClass*)clientData)->CommandHandler (interp, objc, objv);
}



// gmlTcl_EntityDeleteCmdHandler --

extern "C"
void gmlTcl_EntityDeleteCmdHandler (ClientData clientData)
{
  gmlTcl_TEntity*      theEntity = (gmlTcl_TEntity*)clientData;

  theEntity->CommandDeleted ();
}



// gmlTcl_TObjectRenameCallback --

extern "C"
void gmlTcl_EntityRenameCallback (ClientData  clientData,
                                  Tcl_Interp* interp,
                                  char*       oldName,
                                  char*       newName,
                                  int         flags)
{
  ((gmlTcl_TObject*)clientData)->Rename (interp, oldName, newName, flags);
}
